#include <math.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

int main(int argc, char *argv[]) {
  int i, me = 0, nProcs, size, same_length = 1;
  double *local_Buffer;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);
  MPI_Comm_size(MPI_COMM_WORLD, &nProcs);

  while (same_length == 1) {

    // Fill buffer of random length `size`
    srand48(me + (unsigned)time((time_t *)NULL));
    size = (int)ceil(drand48() * 10);
    local_Buffer = (double *)malloc(size * sizeof(double));
    for (i = 0; i < size; i++)
      local_Buffer[i] = (double)-me;

    for (i = 0; i < size; i++)
      printf("[%d] data(%d): %lf\n", me, i, local_Buffer[i]);

    // Determine maximum size
    int result = 0;
    MPI_Allreduce(&size, &result, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);

    int is_it_me = (result == size);
    size = result;
    printf(" [%d]\t\t\t Max size: %d\tme?=%d\n", me, result, is_it_me);

    // Determine how many ranks have maximum size
    MPI_Allreduce(&is_it_me, &result, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
    printf(" [%d]\t\t\t Sum: %d\n", me, result);

    // If only one rank has maximum size
    if (result == 1) {
      same_length = 0;
      if (!is_it_me) {
        // Reallocate to that size
        local_Buffer = (double *)realloc(local_Buffer, size * sizeof(double));
      }

      // Determine which rank it is
      is_it_me = is_it_me ? me : 0;
      MPI_Allreduce(&is_it_me, &result, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
      printf(" [%d]\t\t\t Root: %d\n", me, result);

      // Perform the broadcast
      MPI_Bcast(local_Buffer, size, MPI_DOUBLE, result, MPI_COMM_WORLD);
    } else {
      free(local_Buffer);
    }
  }
  free(local_Buffer);

  MPI_Finalize();
  return 0;
}
