// mpicc Diego-MPI1.c -o Diego-MPI1.x -lm

#include <limits.h>
#include <math.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

int main(int argc, char *argv[]) {
  int i, iter, me, nprocs;
  int size, *all_sizes, same_length;
  double *local_buffer;
  int max, maxproc, nmaxs;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  all_sizes = (int *)malloc(nprocs * sizeof(int));
  same_length = 1;
  iter = 0;
  while (same_length == 1) {
    iter++;
    printf("Iteration: %d\n", iter);
    // Allocate and initialize buffer of random size
    srand48(me + (unsigned)time((time_t *)NULL));
    size = (int)ceil(drand48() * 10);
    local_buffer = (double *)malloc(size * sizeof(double));
    for (i = 0; i < size; i++)
      local_buffer[i] = (double)-me;

    printf("[%d] Size: %d\n", me, size);

    // Gather size of every proc's array
    MPI_Allgather(&size, 1, MPI_INT, all_sizes, 1, MPI_INT, MPI_COMM_WORLD);

    // Calculate max size, which proc holds it, and how many procs do
    max = INT_MIN;
    maxproc = -1;
    nmaxs = 0;
    for (i = 0; i < nprocs; i++) {
      if (all_sizes[i] > max) {
        max = all_sizes[i];
        maxproc = i;
        nmaxs = 1;
      } else if (all_sizes[i] == max)
        nmaxs++;
    }

    // If only one proc holds the max, we can move on
    if (nmaxs == 1)
      same_length = 0;
    // Else, keep trying
    else
      free(local_buffer);
  }

  printf("[%d] Proc with max length is %d (%d)\n", me, maxproc, max);

  // All but max proc allocate buffer of max size
  double *longest_buffer;
  if (me != maxproc)
    longest_buffer = (double *)malloc(sizeof(double) * max);
  else
    longest_buffer = local_buffer;

  // Bcast with root maxproc
  MPI_Bcast(longest_buffer, max, MPI_DOUBLE, maxproc, MPI_COMM_WORLD);

  // compute( longest_buffer );
  for (i = 0; i < max; i++)
    printf("[%d] Longest(%d) = %lf\n", me, i, longest_buffer[i]);

  // Done, deallocate resources
  free(all_sizes);
  free(local_buffer);
  if (me != maxproc)
    free(longest_buffer);

  MPI_Finalize();

  return 0;
}
