// mpicc Diego-MPI2.c -o Diego-MPI2.x

#include <math.h>
#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void f(int *data_in, int *data_out, int n) {
  int i;
  for (i = 0; i < n; i++)
    data_out[i] = data_in[i] + 1;
}

int main(int argc, char *argv[]) {
  int proc_left, proc_right, me, nprocs;
  int i, n = 4, iters = 0, maxiters = 3;
  int *v_now, *v_next, *v_left, *v_right, *fv;
  MPI_Request req_send_left, req_send_right, req_recv_left, req_recv_right;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &me);
  MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

  proc_left = (me + nprocs - 1) % nprocs;
  proc_right = (me + 1) % nprocs;

  v_now = (int *)malloc(n * sizeof(int));
  v_next = (int *)malloc(n * sizeof(int));
  v_left = (int *)malloc(n * sizeof(int));
  v_right = (int *)malloc(n * sizeof(int));
  fv = (int *)malloc(n * sizeof(int));

  // Initialize v_now = me
  for (i = 0; i < n; i++)
    v_now[i] = me;

  while (iters < maxiters) {
    iters++;

    // v_i -> p_{i-1}
    MPI_Isend(v_now, n, MPI_INT, proc_left, proc_left, MPI_COMM_WORLD,
              &req_send_left);
    // f(v_{i-1}) <- p_{i-1}
    MPI_Irecv(v_left, n, MPI_INT, proc_left, proc_left, MPI_COMM_WORLD,
              &req_recv_left);
    // v_{i+1} <- p_{i+1}
    MPI_Irecv(v_right, n, MPI_INT, proc_right, me, MPI_COMM_WORLD,
              &req_recv_right);

    // v_next = f(v_now)
    f(v_now, v_next, n);

    // copy fv and send to next
    memcpy(fv, v_next, n * sizeof(int));
    MPI_Isend(fv, n, MPI_INT, proc_right, me, MPI_COMM_WORLD, &req_send_right);

    // Receive in any order
    MPI_Request reqs[2] = {req_recv_left, req_recv_right};
    int idx;
    MPI_Waitany(2, reqs, &idx, MPI_STATUS_IGNORE);
    while (idx != MPI_UNDEFINED) {
      if (idx == 0) // Received from proc on the left
      {
        // printf("[%d] Received from left(%d): %d\n", me, proc_left,
        // v_left[0]);
        for (i = 0; i < n; i++)
          v_next[i] += v_left[i];
      } else // Received from proc on the right
      {
        // printf("[%d] Received from right(%d): %d\n", me, proc_right,
        // v_right[0]);
        for (i = 0; i < n; i++)
          v_next[i] -= v_right[i];
      }
      MPI_Waitany(2, reqs, &idx, MPI_STATUS_IGNORE);
    }

    // Wait for the sends to complete
    MPI_Wait(&req_send_left, MPI_STATUS_IGNORE);
    MPI_Wait(&req_send_right, MPI_STATUS_IGNORE);

    // Swap v_now and v_next
    int *tmp = v_now;
    v_now = v_next;
    v_next = tmp;
  }

  for (i = 0; i < n; i++)
    printf("[%d] v[%d] = %d\n", me, i, v_now[i]);

  // free the allocated memory
  free(v_now);
  free(v_next);
  free(v_left);
  free(v_right);
  free(fv);

  MPI_Finalize();

  return 0;
}
