#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

int main(int argc, char *argv[]) {
  int me_WORLD, nP_WORLD, root, size, i;
  double * buffer;
  double * buffer_tmp;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &me_WORLD);
  MPI_Comm_size(MPI_COMM_WORLD, &nP_WORLD);

  if (nP_WORLD % 5 != 0 && me_WORLD == 0) {
    printf("[%d] nP is not a multiple of 5 -- try again\n", me_WORLD);
    return -1;
  }
  if (me_WORLD == 0 && argc < 2) {
    printf("specify the root process, please\n");
    return -1;
  }
  root = atoi(argv[1]);
  size = nP_WORLD * 2;

  buffer = (double *)malloc(size * sizeof(double));
  buffer_tmp = (double *)malloc(size * sizeof(double));
  if (me_WORLD == root)
    for (i = 0; i < size; i++)
      buffer[i] = i; //0.2345 * i + 1.0;

  int color, key;
  MPI_Comm COMM_row;
  color = me_WORLD / 5; // We logically map the processes onto a grid of
                        // size r * 5
  key = me_WORLD;       // for ordering purposes

  // procs           ->  me/5
  // 0  1  2  3  4       0 0 0 0 0   -> same communicator
  // 5  6  7  8  9       1 1 1 1 1
  // 10 11 12 13 14      2 2 2 2 2
  // ...                 . . .

  // row communicator
  MPI_Comm_split(MPI_COMM_WORLD, color, key, &COMM_row);

  int me_ROW, nP_ROW;
  MPI_Comm_rank(COMM_row, &me_ROW);
  MPI_Comm_size(COMM_row, &nP_ROW);
  printf("[%d]\tCOMM_row rank: %d / %d\n", me_WORLD, me_ROW, nP_ROW);

  // 0  1  2  3  4       0 1 2 3 4
  // 5  6  7  8  9       0 1 2 3 4
  // 10 11 12 13 14      0 1 2 3 4
  // ...                 . . .

  // column communicator
  MPI_Comm COMM_col;
  int me_COL, nP_COL;
  MPI_Comm_split(MPI_COMM_WORLD, me_WORLD % 5, key, &COMM_col);

  MPI_Comm_rank(COMM_col, &me_COL);
  MPI_Comm_size(COMM_col, &nP_COL);
  printf("[%d]\t\t\tCOMM_col rank: %d / %d\n", me_WORLD, me_COL, nP_COL);

  // scatter by rows
  MPI_Barrier(MPI_COMM_WORLD);
  if (me_WORLD / nP_ROW == root / nP_ROW) {
    int local_root = root % nP_ROW;
    printf("[%d] Scatter from %d.\n", me_WORLD, local_root);
    MPI_Scatter(buffer, size / nP_ROW, MPI_DOUBLE, buffer_tmp, size / nP_ROW, MPI_DOUBLE, local_root, COMM_row);
  }

  // scatter by columns
  MPI_Barrier(MPI_COMM_WORLD);
  int local_root = root / nP_ROW;
  MPI_Scatter(buffer_tmp, size / nP_ROW / nP_COL, MPI_DOUBLE, buffer, size / nP_ROW / nP_COL, MPI_DOUBLE, local_root, COMM_col);
  printf("[%d] %f %f\n", me_WORLD, buffer[0], buffer[1]);

  MPI_Comm_free(&COMM_row);
  MPI_Comm_free(&COMM_col);
  MPI_Finalize();
  return 0;
}
