#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>

int main(int argc, char *argv[]) {
  int me_WORLD, nP_WORLD, root, size, i;
  double *buffer;
  double *buffer_tmp;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &me_WORLD);
  MPI_Comm_size(MPI_COMM_WORLD, &nP_WORLD);

  if (me_WORLD == 0 && argc < 2) {
    printf("specify the root process, please\n");
    return -1;
  }
  root = atoi(argv[1]);
  size = nP_WORLD * 2;

  buffer = (double *)malloc(size * sizeof(double));
  buffer_tmp = (double *)malloc(size * sizeof(double));
  if (me_WORLD == root)
    for (i = 0; i < size; i++)
      buffer_tmp[i] = i;

  int stride = size / nP_WORLD;
  MPI_Scatter(buffer_tmp, stride, MPI_DOUBLE, &buffer[stride * me_WORLD],
              stride, MPI_DOUBLE, root, MPI_COMM_WORLD);

  int left = (me_WORLD - 1 + nP_WORLD) % nP_WORLD;
  int right = (me_WORLD + 1) % nP_WORLD;
  for (i = 0; i < nP_WORLD - 1; i++) {
    int read_pos = (i + me_WORLD) % nP_WORLD;
    int write_pos = (i + me_WORLD + 1) % nP_WORLD;
    /*
0   1   2   3  

0   1   2   3
01  12  23  03
012 123 023 013
0123012301230123

*/
    MPI_Status status;
    MPI_Sendrecv(&buffer[read_pos * stride], stride, MPI_DOUBLE, left, 0,
                 &buffer[write_pos * stride], stride, MPI_DOUBLE, right, 0,
                 MPI_COMM_WORLD, &status);
  }

  printf("[%d] %f %f ... %f\n", me_WORLD, buffer[0], buffer[1],
         buffer[size - 1]);

  MPI_Finalize();
  return 0;
}
